@@ -123,6 +123,7 @@ stop_tcp_listener(Host, Port) ->
 
 tcp_listener_started(IPAddress, Port) ->
     ok = mnesia:dirty_write(
+           rabbit_listener,
            #listener{node = node(),
                      protocol = tcp,
                      host = tcp_host(IPAddress),
@@ -130,19 +131,20 @@ tcp_listener_started(IPAddress, Port) ->
 
 tcp_listener_stopped(IPAddress, Port) ->
     ok = mnesia:dirty_delete_object(
+           rabbit_listener,
            #listener{node = node(),
                      protocol = tcp,
                      host = tcp_host(IPAddress),
                      port = Port}).
 
 active_listeners() ->
-    rabbit_misc:dirty_read_all(listener).
+    rabbit_misc:dirty_read_all(rabbit_listener).
 
 node_listeners(Node) ->
-    mnesia:dirty_read(listener, Node).
+    mnesia:dirty_read(rabbit_listener, Node).
 
 on_node_down(Node) ->
-    ok = mnesia:dirty_delete(listener, Node).
+    ok = mnesia:dirty_delete(rabbit_listener, Node).
 
 start_client(Sock) ->
     {ok, Child} = supervisor:start_child(rabbit_tcp_client_sup, []),
